@@ -9,7 +9,11 @@ const mimetypes = require('./mimetypes');
 const STAGE_KEY = 0x1001;
 const STAGE_VALUE = 0x1002;
 
-let libmime = (module.exports = {
+class Libmime {
+    constructor(config) {
+        this.config = config || {};
+    }
+
     /**
      * Checks if a value is plaintext string (uses only printable 7bit chars)
      *
@@ -22,7 +26,7 @@ let libmime = (module.exports = {
         } else {
             return true;
         }
-    },
+    }
 
     /**
      * Checks if a multi line string containes lines longer than the selected value.
@@ -37,7 +41,7 @@ let libmime = (module.exports = {
      */
     hasLongerLines(str, lineLength) {
         return new RegExp('^.{' + (lineLength + 1) + ',}', 'm').test(str);
-    },
+    }
 
     /**
      * Decodes a string from a format=flowed soft wrapping.
@@ -54,24 +58,24 @@ let libmime = (module.exports = {
                 .split(/\r?\n/)
                 // remove soft linebreaks
                 // soft linebreaks are added after space symbols
-                .reduce((previousValue, currentValue, index) => {
-                    let body = previousValue;
-                    if (delSp) {
-                        // delsp adds spaces to text to be able to fold it
-                        // these spaces can be removed once the text is unfolded
-                        body = body.replace(/[ ]+$/, '');
-                    }
-                    if ((/ $/.test(previousValue) && !/(^|\n)-- $/.test(previousValue)) || index === 1) {
-                        return body + currentValue;
+                .reduce((previousValue, currentValue) => {
+                    if (/ $/.test(previousValue) && !/(^|\n)-- $/.test(previousValue)) {
+                        if (delSp) {
+                            // delsp adds space to text to be able to fold it
+                            // these spaces can be removed once the text is unfolded
+                            return previousValue.slice(0, -1) + currentValue;
+                        } else {
+                            return previousValue + currentValue;
+                        }
                     } else {
-                        return body + '\n' + currentValue;
+                        return previousValue + '\n' + currentValue;
                     }
                 })
                 // remove whitespace stuffing
                 // http://tools.ietf.org/html/rfc3676#section-4.4
                 .replace(/^ /gm, '')
         );
-    },
+    }
 
     /**
      * Adds soft line breaks to content marked with format=flowed to
@@ -87,7 +91,7 @@ let libmime = (module.exports = {
         let flowed = [];
         str.split(/\r?\n/).forEach(line => {
             flowed.push(
-                libmime.foldLines(
+                this.foldLines(
                     line
                         // space stuffing http://tools.ietf.org/html/rfc3676#section-4.2
                         .replace(/^( |From|>)/gim, ' $1'),
@@ -97,7 +101,7 @@ let libmime = (module.exports = {
             );
         });
         return flowed.join('\r\n');
-    },
+    }
 
     /**
      * Encodes a string or an Buffer to an UTF-8 MIME Word (rfc2047)
@@ -108,11 +112,7 @@ let libmime = (module.exports = {
      * @return {String} Single or several mime words joined together
      */
     encodeWord(data, mimeWordEncoding, maxLength) {
-        mimeWordEncoding = (mimeWordEncoding || 'Q')
-            .toString()
-            .toUpperCase()
-            .trim()
-            .charAt(0);
+        mimeWordEncoding = (mimeWordEncoding || 'Q').toString().toUpperCase().trim().charAt(0);
         maxLength = maxLength || 0;
 
         let encodedStr;
@@ -125,10 +125,7 @@ let libmime = (module.exports = {
         if (mimeWordEncoding === 'Q') {
             // https://tools.ietf.org/html/rfc2047#section-5 rule (3)
             encodedStr = libqp.encode(data).replace(/[^a-z0-9!*+\-/=]/gi, chr => {
-                let ord = chr
-                    .charCodeAt(0)
-                    .toString(16)
-                    .toUpperCase();
+                let ord = chr.charCodeAt(0).toString(16).toUpperCase();
                 if (chr === ' ') {
                     return '_';
                 } else {
@@ -137,12 +134,12 @@ let libmime = (module.exports = {
             });
         } else if (mimeWordEncoding === 'B') {
             encodedStr = typeof data === 'string' ? data : libbase64.encode(data);
-            maxLength = maxLength ? Math.max(3, (maxLength - maxLength % 4) / 4 * 3) : 0;
+            maxLength = maxLength ? Math.max(3, ((maxLength - (maxLength % 4)) / 4) * 3) : 0;
         }
 
         if (maxLength && (mimeWordEncoding !== 'B' ? encodedStr : libbase64.encode(data)).length > maxLength) {
             if (mimeWordEncoding === 'Q') {
-                encodedStr = splitMimeEncodedString(encodedStr, maxLength).join('?= =?' + toCharset + '?' + mimeWordEncoding + '?');
+                encodedStr = this.splitMimeEncodedString(encodedStr, maxLength).join('?= =?' + toCharset + '?' + mimeWordEncoding + '?');
             } else {
                 // RFC2047 6.3 (2) states that encoded-word must include an integral number of characters, so no chopping unicode sequences
                 let parts = [];
@@ -174,7 +171,7 @@ let libmime = (module.exports = {
         }
 
         return '=?' + toCharset + '?' + mimeWordEncoding + '?' + encodedStr + (encodedStr.substr(-2) === '?=' ? '' : '?=');
-    },
+    }
 
     /**
      * Decode a complete mime word encoded string
@@ -219,14 +216,19 @@ let libmime = (module.exports = {
             }
             str = Buffer.from(bytes);
         } else if (encoding === 'B') {
-            str = Buffer.from(str, 'base64');
+            str = Buffer.concat(
+                str
+                    .split('=')
+                    .filter(s => s !== '') // filter empty string
+                    .map(str => Buffer.from(str, 'base64'))
+            );
         } else {
             // keep as is, convert Buffer to unicode string, assume utf8
             str = Buffer.from(str);
         }
 
         return libcharset.decode(str, charset);
-    },
+    }
 
     /**
      * Finds word sequences with non ascii text and converts these to mime words
@@ -268,11 +270,11 @@ let libmime = (module.exports = {
 
         encodedValue =
             (startIndex ? decodedValue.substr(0, startIndex) : '') +
-            libmime.encodeWord(decodedValue.substring(startIndex, endIndex), mimeWordEncoding || 'Q', maxLength) +
+            this.encodeWord(decodedValue.substring(startIndex, endIndex), mimeWordEncoding || 'Q', maxLength) +
             (endIndex < decodedValue.length ? decodedValue.substr(endIndex) : '');
 
         return encodedValue;
-    },
+    }
 
     /**
      * Decode a string that might include one or several mime words
@@ -285,7 +287,7 @@ let libmime = (module.exports = {
             (str || '')
                 .toString()
                 // find base64 words that can be joined
-                .replace(/(=\?([^?]+)\?[Bb]\?[^?]+[^^=]\?=)\s*(?==\?([^?]+)\?[Bb]\?[^?]+\?=)/g, (match, left, chLeft, chRight) => {
+                .replace(/(=\?([^?]+)\?[Bb]\?[^?]*\?=)\s*(?==\?([^?]+)\?[Bb]\?[^?]*\?=)/g, (match, left, chLeft, chRight) => {
                     // only mark b64 chunks to be joined if charsets match
                     if (libcharset.normalizeCharset(chLeft || '') === libcharset.normalizeCharset(chRight || '')) {
                         // set a joiner marker
@@ -294,7 +296,7 @@ let libmime = (module.exports = {
                     return match;
                 })
                 // find QP words that can be joined
-                .replace(/(=\?([^?]+)\?[Qq]\?[^?]+\?=)\s*(?==\?([^?]+)\?[Qq]\?[^?]+\?=)/g, (match, left, chLeft, chRight) => {
+                .replace(/(=\?([^?]+)\?[Qq]\?[^?]*\?=)\s*(?==\?([^?]+)\?[Qq]\?[^?]*\?=)/g, (match, left, chLeft, chRight) => {
                     // only mark QP chunks to be joined if charsets match
                     if (libcharset.normalizeCharset(chLeft || '') === libcharset.normalizeCharset(chRight || '')) {
                         // set a joiner marker
@@ -305,18 +307,18 @@ let libmime = (module.exports = {
                 // join base64 encoded words
                 .replace(/(\?=)?__\x00JOIN\x00__(=\?([^?]+)\?[QqBb]\?)?/g, '')
                 // remove spaces between mime encoded words
-                .replace(/(=\?[^?]+\?[QqBb]\?[^?]+\?=)\s+(?==\?[^?]+\?[QqBb]\?[^?]+\?=)/g, '$1')
+                .replace(/(=\?[^?]+\?[QqBb]\?[^?]*\?=)\s+(?==\?[^?]+\?[QqBb]\?[^?]*\?=)/g, '$1')
                 // decode words
-                .replace(/=\?([\w_\-*]+)\?([QqBb])\?([^?]+)\?=/g, (m, charset, encoding, text) => this.decodeWord(charset, encoding, text))
+                .replace(/=\?([\w_\-*]+)\?([QqBb])\?([^?]*)\?=/g, (m, charset, encoding, text) => this.decodeWord(charset, encoding, text))
         );
-    },
+    }
 
     getHex(c) {
         if ((c >= 0x30 /* 0 */ && c <= 0x39) /* 9 */ || (c >= 0x61 /* a */ && c <= 0x66) /* f */ || (c >= 0x41 /* A */ && c <= 0x46) /* F */) {
             return String.fromCharCode(c);
         }
         return false;
-    },
+    }
 
     /**
      * Splits a string by :
@@ -339,7 +341,7 @@ let libmime = (module.exports = {
             key,
             value
         };
-    },
+    }
 
     /**
      * Parses a block of header lines. Does not decode mime words as every
@@ -363,7 +365,7 @@ let libmime = (module.exports = {
         }
 
         for (i = 0, len = lines.length; i < len; i++) {
-            header = libmime.decodeHeader(lines[i]);
+            header = this.decodeHeader(lines[i]);
             if (!headersObj[header.key]) {
                 headersObj[header.key] = [header.value];
             } else {
@@ -372,7 +374,7 @@ let libmime = (module.exports = {
         }
 
         return headersObj;
-    },
+    }
 
     /**
      * Joins parsed header value together as 'value; param1=value1; param2=value2'
@@ -387,8 +389,8 @@ let libmime = (module.exports = {
         Object.keys(structured.params || {}).forEach(param => {
             // filename might include unicode characters so it is a special case
             let value = structured.params[param];
-            if (!libmime.isPlainText(value) || value.length >= 75) {
-                libmime.buildHeaderParam(param, value, 50).forEach(encodedParam => {
+            if (!this.isPlainText(value) || value.length >= 75) {
+                this.buildHeaderParam(param, value, 50).forEach(encodedParam => {
                     if (!/[\s"\\;:/=(),<>@[\]?]|^[-']|'$/.test(encodedParam.value) || encodedParam.key.substr(-1) === '*') {
                         paramsArray.push(encodedParam.key + '=' + encodedParam.value);
                     } else {
@@ -403,7 +405,7 @@ let libmime = (module.exports = {
         });
 
         return structured.value + (paramsArray.length ? '; ' + paramsArray.join('; ') : '');
-    },
+    }
 
     /**
      * Parses a header value with key=value arguments into a structured
@@ -562,7 +564,7 @@ let libmime = (module.exports = {
         });
 
         return response;
-    },
+    }
 
     /**
      * Encodes a string or an Buffer to an UTF-8 Parameter Value Continuation encoding (rfc2231)
@@ -581,7 +583,7 @@ let libmime = (module.exports = {
      */
     buildHeaderParam(key, data, maxLength, fromCharset) {
         let list = [];
-        let encodedStr = typeof data === 'string' ? data : libmime.decode(data, fromCharset);
+        let encodedStr = typeof data === 'string' ? data : this.decode(data, fromCharset);
         let encodedStrArr;
         let chr, ord;
         let line;
@@ -592,7 +594,7 @@ let libmime = (module.exports = {
         maxLength = maxLength || 50;
 
         // process ascii only text
-        if (libmime.isPlainText(data)) {
+        if (this.isPlainText(data)) {
             // check if conversion is even needed
             if (encodedStr.length <= maxLength) {
                 return [
@@ -644,10 +646,10 @@ let libmime = (module.exports = {
                 chr = encodedStr[i];
 
                 if (isEncoded) {
-                    chr = safeEncodeURIComponent(chr);
+                    chr = this.safeEncodeURIComponent(chr);
                 } else {
                     // try to urlencode current char
-                    chr = chr === ' ' ? chr : safeEncodeURIComponent(chr);
+                    chr = chr === ' ' ? chr : this.safeEncodeURIComponent(chr);
                     // By default it is not required to encode a line, the need
                     // only appears when the string contains unicode or special chars
                     // in this case we start processing the line over and encode all chars
@@ -655,7 +657,7 @@ let libmime = (module.exports = {
                         // Check if it is even possible to add the encoded char to the line
                         // If not, there is no reason to use this line, just push it to the list
                         // and start a new line with the char that needs encoding
-                        if ((safeEncodeURIComponent(line) + chr).length >= maxLength) {
+                        if ((this.safeEncodeURIComponent(line) + chr).length >= maxLength) {
                             list.push({
                                 line,
                                 encoded: isEncoded
@@ -677,7 +679,7 @@ let libmime = (module.exports = {
                         line,
                         encoded: isEncoded
                     });
-                    line = chr = encodedStr[i] === ' ' ? ' ' : safeEncodeURIComponent(encodedStr[i]);
+                    line = chr = encodedStr[i] === ' ' ? ' ' : this.safeEncodeURIComponent(encodedStr[i]);
                     if (chr === encodedStr[i]) {
                         isEncoded = false;
                         startPos = i - 1;
@@ -704,7 +706,7 @@ let libmime = (module.exports = {
             key: key + '*' + i + (item.encoded ? '*' : ''),
             value: item.line
         }));
-    },
+    }
 
     /**
      * Returns file extension for a content type string. If no suitable extensions
@@ -714,10 +716,7 @@ let libmime = (module.exports = {
      * @return {String} File extension
      */
     detectExtension(mimeType) {
-        mimeType = (mimeType || '')
-            .toString()
-            .toLowerCase()
-            .replace(/\s/g, '');
+        mimeType = (mimeType || '').toString().toLowerCase().replace(/\s/g, '');
         if (!(mimeType in mimetypes.list)) {
             return 'bin';
         }
@@ -737,7 +736,7 @@ let libmime = (module.exports = {
 
         // use the first one
         return mimetypes.list[mimeType][0] !== '*' ? mimetypes.list[mimeType][0] : 'bin';
-    },
+    }
 
     /**
      * Returns content type for a file extension. If no suitable content types
@@ -747,13 +746,7 @@ let libmime = (module.exports = {
      * @return {String} File extension
      */
     detectMimeType(extension) {
-        extension = (extension || '')
-            .toString()
-            .toLowerCase()
-            .replace(/\s/g, '')
-            .replace(/^\./g, '')
-            .split('.')
-            .pop();
+        extension = (extension || '').toString().toLowerCase().replace(/\s/g, '').replace(/^\./g, '').split('.').pop();
 
         if (!(extension in mimetypes.extensions)) {
             return 'application/octet-stream';
@@ -775,7 +768,7 @@ let libmime = (module.exports = {
 
         // use the first one
         return mimetypes.extensions[extension][0];
-    },
+    }
 
     /**
      * Folds long lines, useful for folding header lines (afterSpace=false) and
@@ -822,89 +815,89 @@ let libmime = (module.exports = {
 
         return result;
     }
-});
 
-/**
- * Splits a mime encoded string. Needed for dividing mime words into smaller chunks
- *
- * @param {String} str Mime encoded string to be split up
- * @param {Number} maxlen Maximum length of characters for one part (minimum 12)
- * @return {Array} Split string
- */
-function splitMimeEncodedString(str, maxlen) {
-    let curLine,
-        match,
-        chr,
-        done,
-        lines = [];
+    /**
+     * Splits a mime encoded string. Needed for dividing mime words into smaller chunks
+     *
+     * @param {String} str Mime encoded string to be split up
+     * @param {Number} maxlen Maximum length of characters for one part (minimum 12)
+     * @return {Array} Split string
+     */
+    splitMimeEncodedString(str, maxlen) {
+        let curLine,
+            match,
+            chr,
+            done,
+            lines = [];
 
-    // require at least 12 symbols to fit possible 4 octet UTF-8 sequences
-    maxlen = Math.max(maxlen || 0, 12);
+        // require at least 12 symbols to fit possible 4 octet UTF-8 sequences
+        maxlen = Math.max(maxlen || 0, 12);
 
-    while (str.length) {
-        curLine = str.substr(0, maxlen);
+        while (str.length) {
+            curLine = str.substr(0, maxlen);
 
-        // move incomplete escaped char back to main
-        if ((match = curLine.match(/[=][0-9A-F]?$/i))) {
-            curLine = curLine.substr(0, match.index);
-        }
+            // move incomplete escaped char back to main
+            if ((match = curLine.match(/[=][0-9A-F]?$/i))) {
+                curLine = curLine.substr(0, match.index);
+            }
 
-        done = false;
-        while (!done) {
-            done = true;
-            // check if not middle of a unicode char sequence
-            if ((match = str.substr(curLine.length).match(/^[=]([0-9A-F]{2})/i))) {
-                chr = parseInt(match[1], 16);
-                // invalid sequence, move one char back anc recheck
-                if (chr < 0xc2 && chr > 0x7f) {
-                    curLine = curLine.substr(0, curLine.length - 3);
-                    done = false;
+            done = false;
+            while (!done) {
+                done = true;
+                // check if not middle of a unicode char sequence
+                if ((match = str.substr(curLine.length).match(/^[=]([0-9A-F]{2})/i))) {
+                    chr = parseInt(match[1], 16);
+                    // invalid sequence, move one char back anc recheck
+                    if (chr < 0xc2 && chr > 0x7f) {
+                        curLine = curLine.substr(0, curLine.length - 3);
+                        done = false;
+                    }
                 }
             }
+
+            if (curLine.length) {
+                lines.push(curLine);
+            }
+            str = str.substr(curLine.length);
         }
 
-        if (curLine.length) {
-            lines.push(curLine);
+        return lines;
+    }
+
+    encodeURICharComponent(chr) {
+        let res = '';
+        let ord = chr.charCodeAt(0).toString(16).toUpperCase();
+
+        if (ord.length % 2) {
+            ord = '0' + ord;
         }
-        str = str.substr(curLine.length);
-    }
 
-    return lines;
-}
-
-function encodeURICharComponent(chr) {
-    let res = '';
-    let ord = chr
-        .charCodeAt(0)
-        .toString(16)
-        .toUpperCase();
-
-    if (ord.length % 2) {
-        ord = '0' + ord;
-    }
-
-    if (ord.length > 2) {
-        for (let i = 0, len = ord.length / 2; i < len; i++) {
-            res += '%' + ord.substr(i, 2);
+        if (ord.length > 2) {
+            for (let i = 0, len = ord.length / 2; i < len; i++) {
+                res += '%' + ord.substr(i, 2);
+            }
+        } else {
+            res += '%' + ord;
         }
-    } else {
-        res += '%' + ord;
+
+        return res;
     }
 
-    return res;
-}
+    safeEncodeURIComponent(str) {
+        str = (str || '').toString();
 
-function safeEncodeURIComponent(str) {
-    str = (str || '').toString();
+        try {
+            // might throw if we try to encode invalid sequences, eg. partial emoji
+            str = encodeURIComponent(str);
+        } catch (E) {
+            // should never run
+            return str.replace(/[^\x00-\x1F *'()<>@,;:\\"[\]?=\u007F-\uFFFF]+/g, '');
+        }
 
-    try {
-        // might throw if we try to encode invalid sequences, eg. partial emoji
-        str = encodeURIComponent(str);
-    } catch (E) {
-        // should never run
-        return str.replace(/[^\x00-\x1F *'()<>@,;:\\"[\]?=\u007F-\uFFFF]+/g, '');
+        // ensure chars that are not handled by encodeURICompent are converted as well
+        return str.replace(/[\x00-\x1F *'()<>@,;:\\"[\]?=\u007F-\uFFFF]/g, chr => this.encodeURICharComponent(chr));
     }
-
-    // ensure chars that are not handled by encodeURICompent are converted as well
-    return str.replace(/[\x00-\x1F *'()<>@,;:\\"[\]?=\u007F-\uFFFF]/g, chr => encodeURICharComponent(chr));
 }
+
+module.exports = new Libmime();
+module.exports.Libmime = Libmime;
