@@ -8,12 +8,90 @@ const ipv6normalize = require('ipv6-normalize');
 const sasl = require('./sasl');
 const crypto = require('crypto');
 const os = require('os');
-const punycode = require('punycode/');
+const punycode = require('punycode.js');
 const EventEmitter = require('events');
 const base32 = require('base32.js');
 
 const SOCKET_TIMEOUT = 60 * 1000;
 
+// Enhanced Status Code mappings based on RFC 3463
+const ENHANCED_STATUS_CODES = {
+    // Success codes (2xx)
+    200: '2.0.0', // System status, or system help reply
+    211: '2.0.0', // System status, or system help reply
+    214: '2.0.0', // Help message
+    220: '2.0.0', // Service ready
+    221: '2.0.0', // Service closing transmission channel
+    235: '2.7.0', // Authentication successful
+    250: '2.0.0', // Requested mail action okay, completed
+    251: '2.1.5', // User not local; will forward
+    252: '2.1.5', // Cannot VRFY user, but will accept message
+    334: '3.7.0', // Server challenge for authentication
+    354: '2.0.0', // Start mail input; end with <CRLF>.<CRLF>
+
+    // Temporary failure codes (4xx)
+    420: '4.4.2', // Timeout or connection lost (non-standard, used by some servers)
+    421: '4.4.2', // Service not available, closing transmission channel
+    450: '4.2.1', // Requested mail action not taken: mailbox unavailable
+    451: '4.3.0', // Requested action aborted: local error in processing
+    452: '4.2.2', // Requested action not taken: insufficient system storage
+    454: '4.7.0', // Temporary authentication failure
+
+    // Permanent failure codes (5xx)
+    500: '5.5.2', // Syntax error, command unrecognized
+    501: '5.5.4', // Syntax error in parameters or arguments
+    502: '5.5.1', // Command not implemented
+    503: '5.5.1', // Bad sequence of commands
+    504: '5.5.4', // Command parameter not implemented
+    521: '5.3.2', // Machine does not accept mail
+    523: '5.3.4', // Message size exceeds server limit (non-standard, used by some servers)
+    530: '5.7.0', // Authentication required
+    535: '5.7.8', // Authentication credentials invalid
+    538: '5.7.0', // Must issue a STARTTLS command first (non-standard)
+    550: '5.1.1', // Requested action not taken: mailbox unavailable
+    551: '5.1.6', // User not local; please try forwarding
+    552: '5.2.2', // Requested mail action aborted: exceeded storage allocation
+    553: '5.1.3', // Requested action not taken: mailbox name not allowed
+    554: '5.6.0', // Transaction failed
+    555: '5.5.4', // MAIL FROM/RCPT TO parameters not recognized or not implemented
+    556: '5.1.10', // RCPT TO syntax error (non-standard)
+    557: '5.7.1', // Delivery not authorized (non-standard, used by some servers)
+    558: '5.2.3' // Message too large for recipient (non-standard, used by some servers)
+};
+
+// Skip enhanced status codes for initial greeting and HELO/EHLO responses
+const SKIPPED_COMMANDS_FOR_ENHANCED_STATUS_CODES = new Set(['HELO', 'EHLO', 'LHLO']);
+
+// Context-specific enhanced status code mappings
+const CONTEXTUAL_STATUS_CODES = {
+    // Mail transaction specific codes
+    MAIL_FROM_OK: '2.1.0', // Originator address valid
+    RCPT_TO_OK: '2.1.5', // Destination address valid
+    DATA_OK: '2.6.0', // Message accepted for delivery
+
+    // Authentication specific codes
+    AUTH_SUCCESS: '2.7.0', // Authentication successful
+    AUTH_REQUIRED: '5.7.0', // Authentication required
+    AUTH_INVALID: '5.7.8', // Authentication credentials invalid
+
+    // Policy specific codes
+    POLICY_VIOLATION: '5.7.1', // Delivery not authorized
+    SPAM_REJECTED: '5.7.1', // Message refused
+
+    // Mailbox specific codes
+    MAILBOX_FULL: '4.2.2', // Mailbox full
+    MAILBOX_NOT_FOUND: '5.1.1', // Mailbox does not exist
+    MAILBOX_SYNTAX_ERROR: '5.1.3', // Invalid mailbox syntax
+
+    // System specific codes
+    SYSTEM_ERROR: '4.3.0', // System error
+    SYSTEM_FULL: '4.3.1', // System storage exceeded
+
+    // Network specific codes
+    NETWORK_ERROR: '4.4.0', // Network routing error
+    CONNECTION_TIMEOUT: '4.4.2' // Connection timeout
+};
+
 /**
  * Creates a handler for new socket
  *
@@ -125,11 +203,13 @@ class SMTPConnection extends EventEmitter {
         this._setListeners(() => {
             // Check that connection limit is not exceeded
             if (this._server.options.maxClients && this._server.connections.size > this._server.options.maxClients) {
-                return this.send(421, this.name + ' Too many connected clients, try again in a moment');
+                return this.send(421, this.name + ' Too many connected clients, try again in a moment', false);
             }
 
             // Keep a small delay for detecting early talkers
-            setTimeout(() => this.connectionReady(), 100);
+            let readyTimer = setTimeout(() => this.connectionReady(), 100);
+            // Unref timer so connection init delay doesn't prevent process exit
+            readyTimer.unref();
         });
     }
 
@@ -188,7 +268,7 @@ class SMTPConnection extends EventEmitter {
                 );
 
                 if (err) {
-                    this.send(err.responseCode || 554, err.message);
+                    this.send(err.responseCode || 554, err.message, false);
                     return this.close();
                 }
 
@@ -204,7 +284,8 @@ class SMTPConnection extends EventEmitter {
                         this.name +
                             ' ' +
                             (this._server.options.lmtp ? 'LMTP' : 'ESMTP') +
-                            (this._server.options.banner ? ' ' + this._server.options.banner : '')
+                            (this._server.options.banner ? ' ' + this._server.options.banner : ''),
+                        false
                     );
 
                     if (typeof next === 'function') {
@@ -229,16 +310,27 @@ class SMTPConnection extends EventEmitter {
             greetingSent = true;
             reverseCb(new Error('Timeout'));
         }, 1500);
+        // Unref timer so DNS timeout doesn't prevent process exit
+        reverseTimer.unref();
+
+        // Helper function to handle resolver results consistently
+        const handleResolverResult = (...args) => {
+            clearTimeout(reverseTimer);
+            if (greetingSent) {
+                return;
+            }
+            greetingSent = true;
+            reverseCb(...args);
+        };
+
         try {
-            // dns.reverse throws on invalid input, see https://github.com/nodejs/node/issues/3112
-            dns.reverse(this.remoteAddress.toString(), (...args) => {
-                clearTimeout(reverseTimer);
-                if (greetingSent) {
-                    return;
-                }
-                greetingSent = true;
-                reverseCb(...args);
-            });
+            // Use custom resolver if provided, otherwise use default dns.reverse
+            if (this._server.options.resolver && typeof this._server.options.resolver.reverse === 'function') {
+                this._server.options.resolver.reverse(this.remoteAddress.toString(), handleResolverResult);
+            } else {
+                // dns.reverse throws on invalid input, see https://github.com/nodejs/node/issues/3112
+                dns.reverse(this.remoteAddress.toString(), handleResolverResult);
+            }
         } catch (E) {
             clearTimeout(reverseTimer);
             if (greetingSent) {
@@ -254,17 +346,33 @@ class SMTPConnection extends EventEmitter {
      *
      * @param {Number} code Response code
      * @param {String|Array} data If data is Array, send a multi-line response
+     * @param {String|Boolean} context Optional context for enhanced status codes
      */
-    send(code, data) {
+    send(code, data, context) {
         let payload;
+        let enhancedCode = this._getEnhancedStatusCode(code, context);
 
         if (Array.isArray(data)) {
-            payload = data.map((line, i, arr) => code + (i < arr.length - 1 ? '-' : ' ') + line).join('\r\n');
+            // Multi-line response - enhanced status code must appear on each line
+            payload = data
+                .map((line, i, arr) => {
+                    let prefix = code + (i < arr.length - 1 ? '-' : ' ');
+                    if (enhancedCode) {
+                        prefix += enhancedCode + ' ';
+                    }
+                    return prefix + line;
+                })
+                .join('\r\n');
         } else {
-            payload = []
-                .concat(code || [])
-                .concat(data || [])
-                .join(' ');
+            // Single line response
+            let parts = [code];
+            if (enhancedCode) {
+                parts.push(enhancedCode);
+            }
+            if (data) {
+                parts.push(data);
+            }
+            payload = parts.join(' ');
         }
 
         if (code >= 400) {
@@ -418,7 +526,7 @@ class SMTPConnection extends EventEmitter {
     }
 
     /**
-     * Checks if a selected command is available and ivokes it
+     * Checks if a selected command is available and invokes it
      *
      * @param {Buffer} command Single line of data from the client
      * @param {Function} callback Callback to run once the command is processed
@@ -437,19 +545,23 @@ class SMTPConnection extends EventEmitter {
         );
 
         let handler;
+        callback = callback || (() => false);
+
+        // If server already closing then ignore commands
+        if (this._server._closeTimeout) {
+            return this.send(421, 'Server shutting down', commandName);
+        }
 
         if (!this._ready) {
             // block spammers that send payloads before server greeting
-            return this.send(421, this.name + ' You talk too soon');
+            return this.send(421, this.name + ' You talk too soon', commandName);
         }
 
         // block malicious web pages that try to make SMTP calls from an AJAX request
         if (/^(OPTIONS|GET|HEAD|POST|PUT|DELETE|TRACE|CONNECT) \/.* HTTP\/\d\.\d$/i.test(command)) {
-            return this.send(421, 'HTTP requests not allowed');
+            return this.send(421, 'HTTP requests not allowed', commandName);
         }
 
-        callback = callback || (() => false);
-
         if (this._upgrading) {
             // ignore any commands before TLS upgrade is finished
             return callback();
@@ -472,7 +584,7 @@ class SMTPConnection extends EventEmitter {
                 switch (commandName) {
                     case 'HELO':
                     case 'EHLO':
-                        this.send(500, 'Error: ' + commandName + ' not allowed in LMTP server');
+                        this.send(500, 'Error: ' + commandName + ' not allowed in LMTP server', false);
                         return setImmediate(callback);
                     case 'LHLO':
                         commandName = 'EHLO';
@@ -488,10 +600,10 @@ class SMTPConnection extends EventEmitter {
             // if the user makes more
             this._unrecognizedCommands++;
             if (this._unrecognizedCommands >= 10) {
-                return this.send(421, 'Error: too many unrecognized commands');
+                return this.send(421, 'Error: too many unrecognized commands', commandName);
             }
 
-            this.send(500, 'Error: command not recognized');
+            this.send(500, 'Error: command not recognized', commandName);
             return setImmediate(callback);
         }
 
@@ -505,7 +617,7 @@ class SMTPConnection extends EventEmitter {
         ) {
             this._unauthenticatedCommands++;
             if (this._unauthenticatedCommands >= this._maxAllowedUnauthenticatedCommands) {
-                return this.send(421, 'Error: too many unauthenticated commands');
+                return this.send(421, 'Error: too many unauthenticated commands', commandName);
             }
         }
 
@@ -537,6 +649,64 @@ class SMTPConnection extends EventEmitter {
         return !this._server.options.disabledCommands.includes(command) && typeof this['handler_' + command] === 'function';
     }
 
+    /**
+     * Determines if enhanced status codes should be used
+     * @returns {Boolean} True if enhanced status codes should be included in responses
+     */
+    _useEnhancedStatusCodes() {
+        return !this._server.options.hideENHANCEDSTATUSCODES;
+    }
+
+    /**
+     * Gets the appropriate enhanced status code for a given SMTP response code and context
+     * @param {Number} code SMTP response code
+     * @param {String|Boolean} context Optional context for more specific status codes
+     * @returns {String} Enhanced status code or empty string if not applicable
+     */
+    _getEnhancedStatusCode(code, context) {
+        if (context === false || !this._useEnhancedStatusCodes()) {
+            return '';
+        }
+
+        // Skip 3xx responses as per RFC 2034
+        if (code >= 300 && code < 400) {
+            return '';
+        }
+
+        // Skip enhanced status codes for initial greeting and HELO/EHLO responses
+        if (context && SKIPPED_COMMANDS_FOR_ENHANCED_STATUS_CODES.has(context)) {
+            return '';
+        }
+
+        // Use contextual codes if available
+        if (context && CONTEXTUAL_STATUS_CODES[context]) {
+            return CONTEXTUAL_STATUS_CODES[context];
+        }
+
+        // Use default mapping
+        if (ENHANCED_STATUS_CODES[code]) {
+            return ENHANCED_STATUS_CODES[code];
+        }
+
+        // 2xx fallback
+        if (code >= 200 && code < 300) {
+            return '2.0.0';
+        }
+
+        // 4xx (transient failure)
+        if (code >= 400 && code < 500) {
+            return '4.0.0';
+        }
+
+        // 5xx (permanent failure)
+        if (code >= 500) {
+            return '5.0.0';
+        }
+
+        // safeguard (non-spec; but should never occur)
+        return '';
+    }
+
     /**
      * Parses commands like MAIL FROM and RCPT TO. Returns an object with the address and optional arguments.
      *
@@ -571,6 +741,11 @@ class SMTPConnection extends EventEmitter {
             let key = part.shift().toUpperCase();
             let value = part.join('=') || true;
 
+            // Skip parameters with empty keys
+            if (!key || key.trim() === '') {
+                return;
+            }
+
             if (typeof value === 'string') {
                 // decode 'xtext'
                 value = value.replace(/\+([0-9A-F]{2})/g, (match, hex) => unescape('%' + hex));
@@ -584,26 +759,53 @@ class SMTPConnection extends EventEmitter {
         });
 
         if (address) {
-            // enforce unycode
+            // Validate email address format
             address = address.split('@');
             if (address.length !== 2 || !address[0] || !address[1]) {
-                // really bad e-mail address validation. was not able to use joi because of the missing unicode support
                 invalid = true;
             } else {
-                try {
-                    address = [address[0] || '', '@', punycode.toUnicode(address[1] || '')].join('');
-                } catch (E) {
-                    this._server.logger.error(
-                        {
-                            tnx: 'punycode',
-                            cid: this.id,
-                            user: (this.session.user && this.session.user.username) || this.session.user
-                        },
-                        'Failed to process punycode domain "%s". error=%s',
-                        address[1],
-                        E.message
-                    );
-                    address = [address[0] || '', '@', address[1] || ''].join('');
+                let localPart = address[0];
+                let domain = address[1];
+
+                // Reasonable length limits to prevent abuse (more permissive than RFC 5321)
+                // Note: Gmail and other providers may exceed RFC 5321's 64-char local-part limit
+                if (localPart.length > 128 || domain.length > 255) {
+                    invalid = true;
+                } else {
+                    // Validate local-part format - permissive validation
+                    // Reject only clearly invalid patterns (dots at edges, consecutive dots)
+                    // Allow common special characters used by email providers
+                    if (localPart.startsWith('.') || localPart.endsWith('.') || localPart.includes('..')) {
+                        invalid = true;
+                    }
+
+                    // Validate domain format (before punycode conversion)
+                    // Reject clearly invalid patterns
+                    if (
+                        !invalid &&
+                        (domain.startsWith('.') || domain.endsWith('.') || domain.includes('..') || domain.includes('.-') || domain.includes('-.') || !/^[a-zA-Z0-9\u0080-\uFFFF.-]+$/.test(domain))
+                    ) {
+                        invalid = true;
+                    }
+                }
+
+                if (!invalid) {
+                    try {
+                        address = [localPart, '@', punycode.toUnicode(domain)].join('');
+                    } catch (E) {
+                        this._server.logger.error(
+                            {
+                                tnx: 'punycode',
+                                cid: this.id,
+                                user: (this.session.user && this.session.user.username) || this.session.user
+                            },
+                            'Failed to process punycode domain "%s". error=%s',
+                            domain,
+                            E.message
+                        );
+                        // If punycode conversion fails, treat as invalid
+                        invalid = true;
+                    }
                 }
             }
         }
@@ -640,9 +842,21 @@ class SMTPConnection extends EventEmitter {
         // reset transaction properties
         session.envelope = {
             mailFrom: false,
-            rcptTo: []
+            rcptTo: [],
+            /** @property {boolean} requireTLS - RFC 8689: Indicates client requires TLS for entire delivery chain */
+            requireTLS: false,
+            /** @property {string} bodyType - RFC 6152: Message body encoding type (7bit or 8bitmime) */
+            bodyType: '7bit',
+            /** @property {boolean} smtpUtf8 - RFC 6531: Indicates UTF-8 support is requested */
+            smtpUtf8: false
         };
 
+        if (!this._server.options.hideDSN)
+            session.envelope.dsn = {
+                ret: null, // RET parameter from MAIL FROM (FULL or HDRS)
+                envid: null // ENVID parameter from MAIL FROM
+            };
+
         session.transaction = this._transactionCounter + 1;
     }
 
@@ -695,13 +909,13 @@ class SMTPConnection extends EventEmitter {
         let hostname = parts[1] || '';
 
         if (parts.length !== 2) {
-            this.send(501, 'Error: syntax: ' + (this._server.options.lmtp ? 'LHLO' : 'EHLO') + ' hostname');
+            this.send(501, 'Error: syntax: ' + (this._server.options.lmtp ? 'LHLO' : 'EHLO') + ' hostname', false);
             return callback();
         }
 
         this.hostNameAppearsAs = hostname.toLowerCase();
 
-        let features = ['PIPELINING', '8BITMIME', 'SMTPUTF8'].filter(feature => !this._server.options['hide' + feature]);
+        let features = ['PIPELINING', '8BITMIME', 'SMTPUTF8', 'ENHANCEDSTATUSCODES', 'DSN'].filter(feature => !this._server.options['hide' + feature]);
 
         if (this._server.options.authMethods.length && this._isSupported('AUTH') && !this.session.user) {
             features.push(['AUTH'].concat(this._server.options.authMethods).join(' '));
@@ -711,6 +925,10 @@ class SMTPConnection extends EventEmitter {
             features.push('STARTTLS');
         }
 
+        if (this.secure && !this._server.options.hideREQUIRETLS) {
+            features.push('REQUIRETLS');
+        }
+
         if (this._server.options.size) {
             features.push('SIZE' + (this._server.options.hideSize ? '' : ' ' + this._server.options.size));
         }
@@ -726,7 +944,14 @@ class SMTPConnection extends EventEmitter {
         }
 
         this._resetSession(); // EHLO is effectively the same as RSET
-        this.send(250, [this.name + ' Nice to meet you, ' + this.clientHostname].concat(features || []));
+
+        // Format HELO response using configured format or default
+        let heloResponse = this._server.options.heloResponse || '%s Nice to meet you, %s';
+        let replacements = [this.name, this.clientHostname];
+        let replacementIndex = 0;
+        let formattedResponse = heloResponse.replace(/%s/g, () => replacements[replacementIndex++] || '');
+
+        this.send(250, [formattedResponse].concat(features || []), false);
 
         callback();
     }
@@ -739,14 +964,21 @@ class SMTPConnection extends EventEmitter {
         let hostname = parts[1] || '';
 
         if (parts.length !== 2) {
-            this.send(501, 'Error: Syntax: HELO hostname');
+            this.send(501, 'Error: Syntax: HELO hostname', false);
             return callback();
         }
 
         this.hostNameAppearsAs = hostname.toLowerCase();
 
         this._resetSession(); // HELO is effectively the same as RSET
-        this.send(250, this.name + ' Nice to meet you, ' + this.clientHostname);
+
+        // Format HELO response using configured format or default
+        let heloResponse = this._server.options.heloResponse || '%s Nice to meet you, %s';
+        let replacements = [this.name, this.clientHostname];
+        let replacementIndex = 0;
+        let formattedResponse = heloResponse.replace(/%s/g, () => replacements[replacementIndex++] || '');
+
+        this.send(250, formattedResponse, false);
 
         callback();
     }
@@ -1107,7 +1339,7 @@ class SMTPConnection extends EventEmitter {
                     this.remotePort = value;
                     break;
                 case 'HELO':
-                    value = Number(value) || 0;
+                    value = (value || '').toString().toLowerCase();
                     this._server.logger.info(
                         {
                             tnx: 'xforward',
@@ -1182,6 +1414,102 @@ class SMTPConnection extends EventEmitter {
         handler(args, callback);
     }
 
+    /**
+     * Validates MAIL FROM parameters
+     * @param {Object} parsed - Parsed address command with args
+     * @returns {Object|null} Returns error object {code, message, enhancedCode} if validation fails, null if successful
+     */
+    _validateMailParams(parsed) {
+        // Validate BODY parameter (RFC 6152 - 8BITMIME)
+        // Note: BINARYMIME is not supported as it requires BDAT command (RFC 3030)
+        if (parsed.args.BODY) {
+            const bodyType = parsed.args.BODY.toLowerCase();
+            if (bodyType !== '7bit' && bodyType !== '8bitmime') {
+                return {
+                    code: 501,
+                    message: 'Invalid BODY parameter value. Must be 7BIT or 8BITMIME',
+                    enhancedCode: null
+                };
+            }
+        }
+
+        // Validate SMTPUTF8 parameter (RFC 6531)
+        if (parsed.args.SMTPUTF8 !== undefined) {
+            if (parsed.args.SMTPUTF8 !== true) {
+                return {
+                    code: 501,
+                    message: 'Invalid SMTPUTF8 parameter. This flag does not accept a value',
+                    enhancedCode: null
+                };
+            }
+        }
+
+        // Validate REQUIRETLS parameter (RFC 8689)
+        if (parsed.args.REQUIRETLS !== undefined) {
+            if (!this.secure) {
+                return {
+                    code: 530,
+                    message: 'REQUIRETLS not permitted on non-TLS connections',
+                    enhancedCode: null
+                };
+            }
+            if (parsed.args.REQUIRETLS !== true) {
+                return {
+                    code: 501,
+                    message: 'Invalid REQUIRETLS parameter. This flag does not accept a value',
+                    enhancedCode: null
+                };
+            }
+        }
+
+        // Validate DSN parameters if DSN is supported
+        if (!this._server.options.hideDSN) {
+            if (parsed.args.RET) {
+                const ret = parsed.args.RET.toUpperCase();
+                if (ret !== 'FULL' && ret !== 'HDRS') {
+                    return {
+                        code: 501,
+                        message: 'Invalid RET parameter value. Must be FULL or HDRS',
+                        enhancedCode: null
+                    };
+                }
+            }
+        }
+
+        return null; // All validations passed
+    }
+
+    /**
+     * Applies validated MAIL FROM parameters to session envelope
+     * @param {Object} parsed - Parsed address command with args
+     */
+    _applyMailParams(parsed) {
+        // Apply BODY parameter
+        if (parsed.args.BODY) {
+            this.session.envelope.bodyType = parsed.args.BODY.toLowerCase();
+        }
+
+        // Apply SMTPUTF8 parameter
+        if (parsed.args.SMTPUTF8 === true) {
+            this.session.envelope.smtpUtf8 = true;
+        }
+
+        // Apply REQUIRETLS parameter
+        if (parsed.args.REQUIRETLS === true) {
+            this.session.envelope.requireTLS = true;
+        }
+
+        // Apply DSN parameters
+        if (!this._server.options.hideDSN) {
+            if (parsed.args.RET) {
+                this.session.envelope.dsn.ret = parsed.args.RET.toUpperCase();
+            }
+            if (parsed.args.ENVID) {
+                this.session.envelope.dsn.envid = parsed.args.ENVID;
+            }
+        }
+    }
+
     /**
      * Processes MAIL FROM command, parses address and extra arguments
      */
@@ -1193,7 +1521,7 @@ class SMTPConnection extends EventEmitter {
 
         // sender address can be empty, so we only check if parsing failed or not
         if (!parsed) {
-            this.send(501, 'Error: Bad sender address syntax');
+            this.send(501, 'Error: Bad sender address syntax', 'MAILBOX_SYNTAX_ERROR');
             return callback();
         }
 
@@ -1203,10 +1531,21 @@ class SMTPConnection extends EventEmitter {
         }
 
         if (!this._server.options.hideSize && this._server.options.size && parsed.args.SIZE && Number(parsed.args.SIZE) > this._server.options.size) {
-            this.send(552, 'Error: message exceeds fixed maximum message size ' + this._server.options.size);
+            this.send(552, 'Error: message exceeds fixed maximum message size ' + this._server.options.size, 'SYSTEM_FULL');
             return callback();
         }
 
+        // Validate all MAIL FROM parameters
+        const validationError = this._validateMailParams(parsed);
+        if (validationError) {
+            this.send(validationError.code, validationError.message, validationError.enhancedCode);
+            return callback();
+        }
+
+        // Apply validated parameters to session envelope before calling onMailFrom
+        // so the handler can access them
+        this._applyMailParams(parsed);
+
         this._server.onMailFrom(parsed, this.session, err => {
             if (err) {
                 this.send(err.responseCode || 550, err.message);
@@ -1215,7 +1554,7 @@ class SMTPConnection extends EventEmitter {
 
             this.session.envelope.mailFrom = parsed;
 
-            this.send(250, 'Accepted');
+            this.send(250, 'Accepted', 'MAIL_FROM_OK');
             callback();
         });
     }
@@ -1228,7 +1567,7 @@ class SMTPConnection extends EventEmitter {
 
         // recipient address can not be empty
         if (!parsed || !parsed.address) {
-            this.send(501, 'Error: Bad recipient address syntax');
+            this.send(501, 'Error: Bad recipient address syntax', 'MAILBOX_SYNTAX_ERROR');
             return callback();
         }
 
@@ -1237,6 +1576,38 @@ class SMTPConnection extends EventEmitter {
             return callback();
         }
 
+        // Process DSN parameters if DSN is supported
+        if (!this._server.options.hideDSN) {
+            // Validate NOTIFY parameter
+            if (parsed.args.NOTIFY) {
+                const notify = parsed.args.NOTIFY.toUpperCase();
+                const validNotifyValues = ['NEVER', 'SUCCESS', 'FAILURE', 'DELAY'];
+                const notifyValues = notify.split(',');
+
+                for (let value of notifyValues) {
+                    if (!validNotifyValues.includes(value)) {
+                        this.send(501, 'Error: NOTIFY parameter must be NEVER, SUCCESS, FAILURE, or DELAY');
+                        return callback();
+                    }
+                }
+
+                // NEVER cannot be combined with other values
+                if (notifyValues.includes('NEVER') && notifyValues.length > 1) {
+                    this.send(501, 'Error: NOTIFY=NEVER cannot be combined with other values');
+                    return callback();
+                }
+
+                parsed.dsn = parsed.dsn || {};
+                parsed.dsn.notify = notifyValues;
+            }
+
+            // Store ORCPT parameter
+            if (parsed.args.ORCPT) {
+                parsed.dsn = parsed.dsn || {};
+                parsed.dsn.orcpt = parsed.args.ORCPT;
+            }
+        }
+
         this._server.onRcptTo(parsed, this.session, err => {
             if (err) {
                 this.send(err.responseCode || 550, err.message);
@@ -1256,7 +1627,7 @@ class SMTPConnection extends EventEmitter {
                 this.session.envelope.rcptTo.push(parsed);
             }
 
-            this.send(250, 'Accepted');
+            this.send(250, 'Accepted', 'RCPT_TO_OK');
             callback();
         });
     }
@@ -1310,17 +1681,17 @@ class SMTPConnection extends EventEmitter {
                     if (/Error\]$/i.test(Object.prototype.toString.call(response))) {
                         this.send(response.responseCode || 450, response.message);
                     } else {
-                        this.send(250, typeof response === 'string' ? response : 'OK: message accepted');
+                        this.send(250, typeof response === 'string' ? response : 'OK: message accepted', 'DATA_OK');
                     }
                 });
             } else if (this._server.options.lmtp) {
                 // separate success response for every recipient when using LMTP
                 for (i = 0, len = this.session.envelope.rcptTo.length; i < len; i++) {
-                    this.send(250, typeof message === 'string' ? message : 'OK: message accepted');
+                    this.send(250, typeof message === 'string' ? message : 'OK: message accepted', 'DATA_OK');
                 }
             } else {
                 // single success response when using SMTP
-                this.send(250, typeof message === 'string' ? message : 'OK: message queued');
+                this.send(250, typeof message === 'string' ? message : 'OK: message queued', 'DATA_OK');
             }
 
             this._transactionCounter++;
@@ -1461,7 +1832,7 @@ class SMTPConnection extends EventEmitter {
                     user: (this.session.user && this.session.user.username) || this.session.user,
                     cipher
                 },
-                'Connection upgraded to TLS using ',
+                'Connection upgraded to TLS using',
                 cipher || 'N/A'
             );
             this._server.onSecure(this._socket, this.session, err => {
