@@ -6,7 +6,7 @@ const SMTPConnection = require('./smtp-connection').SMTPConnection;
 const tlsOptions = require('./tls-options');
 const EventEmitter = require('events');
 const shared = require('nodemailer/lib/shared');
-const punycode = require('punycode/');
+const punycode = require('punycode.js');
 const crypto = require('crypto');
 const base32 = require('base32.js');
 
@@ -36,6 +36,21 @@ class SMTPServer extends EventEmitter {
             this.options.authMethods = ['LOGIN', 'PLAIN'];
         }
 
+        // set default value for hideENHANCEDSTATUSCODES to true (disable enhanced status codes by default)
+        if (this.options.hideENHANCEDSTATUSCODES === undefined) {
+            this.options.hideENHANCEDSTATUSCODES = true;
+        }
+
+        // set default value for hideDSN to true (disable delivery status notifications by default)
+        if (this.options.hideDSN === undefined) {
+            this.options.hideDSN = true;
+        }
+
+        // set default value for hideREQUIRETLS to true (disable REQUIRETLS by default, opt-in)
+        if (this.options.hideREQUIRETLS === undefined) {
+            this.options.hideREQUIRETLS = true;
+        }
+
         this.logger = shared.getLogger(this.options, {
             component: this.options.component || 'smtp-server'
         });
@@ -157,6 +172,9 @@ class SMTPServer extends EventEmitter {
                 return realCallback();
             }
         }, timeout);
+
+        // Unref timer so it doesn't prevent process exit if all connections close naturally
+        this._closeTimeout.unref();
     }
 
     /**
@@ -370,7 +388,7 @@ class SMTPServer extends EventEmitter {
                         if (commandName !== 'PROXY') {
                             try {
                                 socket.end('* BAD Invalid PROXY header\r\n');
-                            } catch (E) {
+                            } catch {
                                 // ignore
                             }
                             return;
@@ -381,19 +399,24 @@ class SMTPServer extends EventEmitter {
 
                             socketOptions.ignore = this.options.ignoredHosts && this.options.ignoredHosts.includes(socketOptions.remoteAddress);
 
-                            if (!socketOptions.ignore) {
-                                this.logger.info(
-                                    {
-                                        tnx: 'proxy',
-                                        cid: socketOptions.id,
-                                        proxy: params[1].trim().toLowerCase()
-                                    },
-                                    '[%s] PROXY from %s through %s (%s)',
-                                    socketOptions.id,
-                                    params[1].trim().toLowerCase(),
-                                    params[2].trim().toLowerCase(),
-                                    JSON.stringify(params)
-                                );
+                            try {
+                                if (!socketOptions.ignore) {
+                                    this.logger.info(
+                                        {
+                                            tnx: 'proxy',
+                                            cid: socketOptions.id,
+                                            proxy: params[1].trim().toLowerCase()
+                                        },
+                                        '[%s] PROXY from %s through %s (%s)',
+                                        socketOptions.id,
+                                        params[1].trim().toLowerCase(),
+                                        params[2].trim().toLowerCase(),
+                                        JSON.stringify(params)
+                                    );
+                                }
+                            } catch {
+                                socket.end('* BAD Invalid PROXY header\r\n');
+                                return;
                             }
 
                             if (params[3]) {
@@ -497,7 +520,7 @@ class SMTPServer extends EventEmitter {
             if (returned) {
                 try {
                     tlsSocket.end();
-                } catch (E) {
+                } catch {
                     //
                 }
                 return;
